{% extends "!layout.html" %}
<link rel="canonical" href="{{ theme_canonical_url }}{{ pagename }}.html" />

{% block extrahead %}
{% if release == "master" %}
<!--
  Search engines should not index the master version of documentation.
  Stable documentation are built without release == 'master'.
-->
<meta name="robots" content="noindex">
{% endif %}
{{ super() }}
{% endblock %}

{% block menu %}
{% if release == "master" %}
<div>
  <a style="color:#F05732" href="{{ theme_canonical_url }}{{ pagename }}.html">
    You are viewing unstable developer preview docs.
    Click here to view docs for latest stable release.
  </a>
</div>
{% endif %}
{{ super() }}
{% endblock %}

{% block sidebartitle %}
    <div class="version">
      <a href='https://pytorch.org/docs/versions.html'>{{ version }} &#x25BC</a>
    </div>
    {% include "searchbox.html" %}
{% endblock %}


{% block footer %}
{{ super() }}
<script script type="text/javascript">
  var collapsedSections = ['Developer Notes', 'Language Bindings', 'Libraries', 'Community'];
</script>

<img height="1" width="1" style="border-style:none;" alt="" src="https://www.googleadservices.com/pagead/conversion/795629140/?label=txkmCPmdtosBENSssfsC&amp;guid=ON&amp;script=0"/>
{% endblock %}
