/* Auto-generated by generate-wrappers.py script. Do not modify */

#if defined(__i386__) || defined(__i686__) || defined(__x86_64__)
#include <q8gavgpool/up8x7-sse2.c>
#endif /* defined(__i386__) || defined(__i686__) || defined(__x86_64__) */
