/* Auto-generated by generate-wrappers.py script. Do not modify */

#if defined(__i386__) || defined(__i686__) || defined(__x86_64__)
#include <x8zip/x2-sse2.c>
#endif /* defined(__i386__) || defined(__i686__) || defined(__x86_64__) */
