/* Auto-generated by generate-wrappers.py script. Do not modify */

#if defined(__arm__) || defined(__aarch64__)
#include <q8gavgpool/mp8x7p7q-neon.c>
#endif /* defined(__arm__) || defined(__aarch64__) */
