
#error This is used exclusively for generating the documentation (not a real header)

//! \namespace torch::jit::fuser
//! \brief Main PyTorch JIT Fuser namespace

//! \namespace torch::jit::fuser::cuda
//! \brief CUDA specific components

//! \namespace torch::jit::fuser::cuda::executor_utils
//! \brief Fuser executor related utilities

//! \namespace torch::jit::fuser::kir
//! \brief Kernel IR

//! \namespace torch::jit::fuser::ir_utils
//! \brief IR manipulation utilities

//! \namespace torch::jit::fuser::loop_utils
//! \brief Loop utilities

//! \namespace torch::jit::fuser::scope_utils
//! \brief Scope utilities
