.codeblock-height-limiter {
    max-height: 500px;
    overflow: scroll;
}

.highlight pre {
  border: 1px solid rgba(0,0,0,0.15);
  border-radius: 4px;
}

div.cell div.cell_input {
  border-bottom-width: 0px;
}
