macro - CONV_ATTRS
- KernelShape : vector<int>
- Pads : vector<int> : {0, 0}
- Strides : vector<int> : {1, 1}
- Group : int : 1
- Dilations: vector<int> : {1, 1}
endmacro

macro - POOL_ATTRS
- KernelShape : vector<int>
- Pads : vector<int> : {0, 0}
- Strides : vector<int> : {1, 1}
endmacro

Relu

Conv
- CONV_ATTRS

ConvRelu : Conv
- CONV_ATTRS

ConvTranspose
- CONV_ATTRS

AveragePool
- POOL_ATTRS

AveragePoolRelu : AveragePool
- POOL_ATTRS

MaxPool
- POOL_ATTRS

MaxPoolRelu : MaxPool
- POOL_ATTRS

Sum

SumRelu : Sum

Send
- Destination : string

Receive
- Source : string

BatchNormalization
- Epsilon : float : 1e-5f
- Momentum : float : 0.9f
- Spatial : bool : true
- IsTest : bool : false

Clip
- Min : float
- Max : float

FC
- Axis : int : 1
- AxisW : int : 1

GivenTensorFill
Concat
- Axis : int : -1
- AddAxis : bool : false
Softmax
ChannelShuffle
Add
- Broadcast : int : 0
Reshape
Flatten

CopyToOpenCL
CopyFromOpenCL

NCHW2NHWC
NHWC2NCHW

Declare
Export
